"""
Functions to read spectral response function data with pyspectral
"""

import numpy as np
from pyspectral.rsr_reader import RelativeSpectralResponse
from typing import Union, List, Tuple, Iterator, Optional


"""___Authorship___"""
__author__ = "Sam Hunt"
__created__ = "5/11/2020"


def return_band_names(
    platform_name: str,
    sensor_name: str,
    band_names: Optional[List[str]] = None,
    min_wl: Optional[float] = None,
    max_wl: Optional[float] = None,
) -> List[str]:
    """
    Returns band names for specified sensor from `pyspectral <https://pyspectral.readthedocs.io/en/master/installation.html#static-data>`_ library.

    :param platform_name: satellite name
    :param sensor_name: name of instrument on satellite
    :param band_names: (optional) if omitted all sensor band names are returned, otherwise submitted band names validated and returned
    :param min_wl: minimum wavelength to include in range
    :param max_wl: maximum wavelength to include in range

    :return: band names
    """

    srf_util = SensorSRFUtil(platform_name, sensor_name)
    return srf_util.return_band_names(
        band_names=band_names, min_wl=min_wl, max_wl=max_wl
    )


def return_band_centres(
    platform_name: str,
    sensor_name: str,
    band_names: Optional[List[str]] = None,
    detector_name: Optional[str] = None,
    min_wl: Optional[float] = None,
    max_wl: Optional[float] = None,
) -> np.ndarray:
    """
    Returns band centres for specified sensor from `pyspectral <https://pyspectral.readthedocs.io/en/master/installation.html#static-data>`_ library.

    :param platform_name: satellite name
    :param sensor_name: name of instrument on satellite
    :param band_names: name of bands to return band centres of, if omitted all band returned
    :param detector_name: name of sensor detector. Can be used in sensor has SRF data for for different
    detectors separately - if not specified in this case different
    :param min_wl: minimum wavelength to include in range
    :param max_wl: maximum wavelength to include in range

    :return: band centres in nm
    """

    srf_util = SensorSRFUtil(
        platform_name, sensor_name, detector_name, band_names=band_names
    )
    return srf_util.return_band_centres(min_wl=min_wl, max_wl=max_wl)


def return_srf(
    platform_name: str,
    sensor_name: str,
    band_name: str = None,
    detector_name: Union[None, str] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """

    Returns srf data for named band of for specified sensor from `pyspectral <https://pyspectral.readthedocs.io/en/master/installation.html#static-data>`_ library.

    :param platform_name: satellite name
    :param sensor_name: name of instrument on satellite
    :param band_name: name of sensor band
    :param detector_name: (optional) name of sensor detector. Can be used in sensor has SRF data for for different
    detectors separately - if not specified in this case different

    :return: band srf
    :return: band srf wavelength coordinates
    """

    srf_util = SensorSRFUtil(platform_name, sensor_name, detector_name)
    return srf_util.return_srf(band_name)


def return_iter_srf(
    platform_name: str,
    sensor_name: str,
    band_names: Optional[List[str]] = None,
    detector_name: Optional[str] = None,
) -> Iterator:
    """
    Returns iterable of band srfs for specified sensor from `pyspectral <https://pyspectral.readthedocs.io/en/master/installation.html#static-data>`_ library.

    :param platform_name: satellite name
    :param sensor_name: name of instrument on satellite
    :param band_names: name of bands to iterate through, if omitted all bands included
    :param detector_name: name of sensor detector. Can be used in sensor has SRF data for for different
    detectors separately - if not specified in this case different

    :return: iterable that returns band srf and srf wavelength coordinates at each iteration
    """

    srf_util = SensorSRFUtil(
        platform_name, sensor_name, detector_name, band_names=band_names
    )
    return iter(srf_util)


class SensorSRFUtil:
    """
    Helper class to define repeating functions along a coordinate axis

    from `pyspectral <https://pyspectral.readthedocs.io/en/master/installation.html#static-data>`_ library.

    :param platform_name: satellite name
    :param sensor_name: name of instrument on satellite
    :param detector_name: (optional) name of sensor detector. Can be used in sensor has SRF data for for different
    detectors separately - if not specified in this case different
    :param band_names: (optional) sensor bands to evaluate band integral for, if omitted band integral evaluated for
    all bands within spectral range of datar
    """

    def __init__(
        self,
        platform_name,
        sensor_name,
        detector_name: Union[None, str] = "det-1",
        band_names: Union[None, List[str]] = None,
    ):

        # Set attributes from arguments
        self.sensor = RelativeSpectralResponse(platform_name, sensor_name)
        self.detector_name = "det-1" if detector_name is None else detector_name

        # Unpack and validate selected bands
        self.band_names = self.return_band_names(band_names)
        self.band_centres = self.return_band_centres()

    def return_band_names(
        self,
        band_names: Optional[str] = None,
        min_wl: Optional[float] = None,
        max_wl: Optional[float] = None,
    ) -> List[str]:
        """
        Returns band names for specified sensor bands

        :param band_names: if omitted all sensor band names are returned,
         otherwise submitted band names validated and returned
        :param min_wl: minimum wavelength to include in range
        :param max_wl: maximum wavelength to include in range

        :return: band names
        """

        sensor_band_names = self.return_sensor_band_names()
        sensor_band_centres = self.return_band_centres()

        selected_band_centres = self.return_band_centres(min_wl=min_wl, max_wl=max_wl)

        selected_band_names = [
            b
            for b, c in zip(sensor_band_names, sensor_band_centres)
            if c in selected_band_centres
        ]

        if band_names is None:
            band_names = selected_band_names
        else:
            if not set(band_names).issubset(set(selected_band_names)):
                raise ValueError(
                    "band names must be one of - " + str(selected_band_names)
                )
            band_names = band_names

        return band_names

    def return_band_centres(
        self, min_wl: Optional[float] = None, max_wl: Optional[float] = None
    ) -> np.ndarray:
        """
        Returns band centres for specified sensor bands

        :param min_wl: minimum wavelength to include in range
        :param max_wl: maximum wavelength to include in range

        :return: band centres in nm
        """

        band_names = self.return_sensor_band_names()

        band_centres = [
            self.sensor.rsr[band_name][self.detector_name]["central_wavelength"]
            * self.sensor.si_scale
            / 1e-9
            for band_name in band_names
        ]

        if min_wl is not None:
            band_centres = [b for b in band_centres if (b > min_wl)]

        if max_wl is not None:
            band_centres = [b for b in band_centres if (b < max_wl)]

        band_centres = np.array(band_centres)

        return band_centres

    def return_sensor_band_names(self) -> List[str]:
        """
        Returns list of all sensor band names

        :return: sensor band names
        """

        return list(self.sensor.rsr.keys())

    def return_srf(self, band_name: str) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns srf data for specified sensor band

        :param band_name: sensor band name

        :return: band srf
        :return: band srf wavelength coordinates
        """

        srf = self.sensor.rsr[band_name][self.detector_name][
            "response"
        ]  # gets rsr for given band
        wl_srf = 1000 * self.sensor.rsr[band_name][self.detector_name]["wavelength"]
        return srf, wl_srf

    def __iter__(self):

        # Define counter
        self.i = 0
        return self

    def __next__(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns ith function

        :return: band srf
        :return: band srf wavelength coordinates
        """

        # Iterate through bands
        if self.i < len(self.band_names):

            # Update counter
            self.i += 1

            return self.return_srf(self.band_names[self.i - 1])

        else:
            raise StopIteration


if __name__ == "__main__":
    pass



"""matheo - Matheo is a python package with mathematical algorithms for use in earth observation data and tools."""

__author__ = "Sam Hunt, Pieter De Vis <kavya.jagan@npl.co.uk>"
__all__ = []

from ._version import __version__



"""
Functions to define commonly used function forms
"""

import numpy as np
from scipy.stats import norm
from typing import Union, Callable, Iterator, Tuple


__author__ = "Sam Hunt"
__created__ = "27/10/2020"


def f_tophat(x: np.ndarray, centre: float, width: float) -> np.ndarray:
    """
    Evaluate a top hat function along x

    :param x: coordinate array
    :param centre: x position of centre top hat
    :param width: FWHM width of top hat

    :return: top hat function
    """

    y = np.ones(x.shape)
    y[x < centre - width / 2] = 0
    y[x > centre + width / 2] = 0

    return y


def f_gaussian(x: np.ndarray, centre: float, width: float) -> np.ndarray:
    """
    Evaluate a gaussian function along x

    :param x: coordinate array
    :param centre: mean of distribution
    :param width: FWHM of distribution

    :return: gaussian function
    """
    std = width / (2 * np.sqrt(2 * np.log(2)))
    return norm.pdf(x, centre, std) * np.sqrt(2 * np.pi) * std


def f_triangle(x: np.ndarray, centre: float, width: float) -> np.ndarray:
    """
    Evaluate a top hat function along x

    :param x: coordinate array
    :param centre: x position of start of triangle
    :param width: half width of base of triangle (i.e. FWHM)

    :return: top hat function
    :rtype: numpy.ndarray
    """

    y = np.zeros(x.shape)

    start = centre - width
    stop = centre + width

    first_half = np.logical_and(start < x, x <= centre)
    y[first_half] = (x[first_half] - start) / (centre - start)

    second_half = np.logical_and(centre < x, x < stop)
    y[second_half] = (stop - x[second_half]) / (stop - centre)

    return y


def f_normalised(f: Callable, x: np.ndarray, *args, high_res_sampling: float = 0.001):
    """
    Return normalised f along x

    :param f: defining function
    :param x: coordinate array
    :param args: f parameters
    :param high_res_sampling: (default 0.001) sampling to evaluate area of function to normalise by

    :return: normalised function
    """

    y = f(x, *args)

    x_highres = np.arange(min(x), max(x), high_res_sampling)
    y_highres = f(x_highres, *args)

    y /= np.trapz(y_highres, x_highres)

    return y


def repeat_f(
    f: Callable,
    centres: np.ndarray,
    widths: np.ndarray,
    normalise: bool = False,
    x_sampling: float = 0.01,
    xlim_width: float = 3.0,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Evaluates repeating functions along a coordinate axis

    .. note::
        Defines all function on common x coordinates, so may use a lot of memory if many bands defined with x sampling.
        For a lower memory solution try **matheo.utils.function_def.iter_fs()**..

    :param f: function to repeat
    :param centres: function centres
    :param widths: function widths
    :param normalise: (default True) switch to define if area of return SRFs should be normalised to 1
    :param x_sampling: sampling along function coordinates
    :param xlim_width: (default 3) multiple function widths to define function over

    :return: evaluated functions
    :return: evaluated function coordinates
    """

    fs = RepeatingFuncUtil(
        f=f,
        centres=centres,
        widths=widths,
        normalise=normalise,
        x_sampling=x_sampling,
        xlim_width=xlim_width,
    )

    return fs.return_fs()


def iter_f(
    f: Callable,
    centres: np.ndarray,
    widths: np.ndarray,
    normalise: bool = False,
    x_sampling: float = 0.01,
    xlim_width: float = 3.0,
) -> Iterator:
    """
    Returns iterator to evaluate repeating functions along a coordinate axis

    .. note::
        Offers a lower memory solution to **matheo.utils.function_def.return_fs()**.

    :param f: function to repeat
    :param centres: distribution centres
    :param widths: distribution widths
    :param normalise: (default True) switch to define if area of return SRFs should be normalised to 1
    :param x_sampling: sampling along function coordinates
    :param xlim_width: (default 3) multiple function widths to define function over

    :return: repeating function iterator
    """

    fs = RepeatingFuncUtil(
        f=f,
        centres=centres,
        widths=widths,
        normalise=normalise,
        x_sampling=x_sampling,
        xlim_width=xlim_width,
    )

    return iter(fs)


class RepeatingFuncUtil:
    """
    Helper class to define repeating functions along a coordinate axis

    :param f: function to repeat
    :param centres: distribution centres
    :param widths: distribution widths
    :param normalise: (default True) switch to define if area of return SRFs should be normalised to 1
    :param x_sampling: sampling along function coordinates
    :param xlim_width: (default 3) multiple function widths to define function over
    """

    def __init__(
        self,
        f: Callable,
        centres: np.ndarray,
        widths: np.ndarray,
        normalise: bool = False,
        x_sampling: float = 0.01,
        xlim_width: float = 3.0,
    ):

        # Set attributes from arguments
        self.f = f
        self.centres = centres
        self.widths = widths
        self.normalise = normalise
        self.x_sampling = x_sampling
        self.xlim_width = xlim_width

    def return_fs(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluates all repeating functions along a coordinate axis

        :return: evaluated functions
        :return: evaluated function coordinates
        """

        x = np.arange(
            min(self.centres - self.xlim_width * self.widths),
            max(self.centres + self.xlim_width * self.widths + 1),
            self.x_sampling,
        )

        # Evaluate function
        ys = np.zeros((len(self.centres), len(x)))
        for i_band, (centre, width) in enumerate(zip(self.centres, self.widths)):

            if self.normalise:
                ys[i_band, :] = f_normalised(self.f, x, centre, width)
            else:
                ys[i_band, :] = self.f(x, centre, width)

        return ys, x

    def __iter__(self):

        # Define counter
        self.i = 0
        return self

    def __next__(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns ith function

        :return: evaluated function
        :return: evaluated function coordinates
        """

        # Iterate through bands
        if self.i < len(self.centres):

            centre = self.centres[self.i]
            width = self.widths[self.i]

            x = np.arange(
                centre - self.xlim_width * width,
                centre + self.xlim_width * width + 1,
                self.x_sampling,
            )

            if self.normalise:
                y = f_normalised(self.f, x, centre, width)
            else:
                y = self.f(x, centre, width)

            # Update counter
            self.i += 1

            return y, x

        else:
            raise StopIteration


if __name__ == "__main__":
    pass



"""
Tests for punpy_util module
"""

from matheo.utils.punpy_util import _max_dim, _unc_to_dim, func_with_unc
import unittest
import numpy as np

"""___Authorship___"""
__author__ = "Sam Hunt"
__created__ = "1/8/2021"


class TestPunpyUtils(unittest.TestCase):

    def test__max_dim(self):
        a = np.zeros((3, 4, 5, 6))
        b = np.zeros(3)

        self.assertEqual(_max_dim([a, b]), 4)

    def test__unc_to_dim_None(self):
        self.assertIsNone(_unc_to_dim(None, 2))

    def test__unc_to_dim_01_xlen(self):
        np.testing.assert_array_equal(_unc_to_dim(3, 1, x_len=2), np.array([3, 3]))

    def test__unc_to_dim_01_x(self):
        np.testing.assert_array_equal(
            _unc_to_dim(0.1, 1, x=np.array([10, 20])), np.array([1, 2])
        )

    def test__unc_to_dim_02_xlen(self):
        np.testing.assert_array_equal(
            _unc_to_dim(3, 2, x_len=2), np.array([[3, 0], [0, 3]])
        )

    def test__unc_to_dim_02_x(self):
        np.testing.assert_array_equal(
            _unc_to_dim(0.1, 2, x=np.array([10, 20])), np.array([[1, 0], [0, 2]])
        )

    def test__unc_to_dim_12_x(self):
        np.testing.assert_array_equal(
            _unc_to_dim(np.array([1, 2]), 2), np.array([[1, 0], [0, 2]])
        )

    def test_func_with_unc_None(self):
        def poly(x1, x2):
            return x1**2 + 3 * x2

        y, u_y = func_with_unc(
            poly,
            params={"x1": np.array([1, 2, 3]), "x2": np.array([1, 2, 3])},
            u_params={"x1": None, "x2": None},
        )

        np.testing.assert_array_equal(y, np.array([4, 10, 18]))
        self.assertIsNone(u_y)

    def test_func_with_unc(self):
        def poly(x1, x2, x3):
            return x1**2 + 3 * x2 + x3

        y, u_y = func_with_unc(
            poly,
            params={"x1": np.array([1, 2, 3]), "x2": np.array([1, 2, 3]), "x3": 1},
            u_params={"x1": 0.1, "x2": np.array([3, 3, 3])},
            parallel=False,  # have to test punpy not in parallel for some reason...
        )

        np.testing.assert_array_equal(y, np.array([5, 11, 19]))
        # np.testing.assert_array_almost_equal(u_y, np.array([7.60914338, 9.31506956, 9.72720938]))


if __name__ == "__main__":
    unittest.main()



"""
Functions to band integrate spectra for given spectral response function.
"""

from punpy import MCPropagation
import numpy as np
from typing import Union, Callable, Tuple, Iterable, Dict
import multiprocessing


__author__ = "Sam Hunt"
__created__ = "24/7/2020"


def _max_dim(arrays: Iterable[np.ndarray]) -> int:
    """
    Return max dimension of input numpy arrays

    :param arrays: n input numpy arrays
    :return: maximum dimension
    """

    dims = []
    for array in arrays:
        dims.append(np.ndim(array))

    return max(dims)


def _unc_to_dim(
    unc: Union[float, np.ndarray], dim: int, x: np.ndarray = None, x_len: int = None
) -> np.ndarray:
    """
    Scales up uncertainty to given dimension (e.g. float to full vector, vector to diagonal maxtrix)

    :param unc: uncertainty value (in percent) or vector
    :param dim: target dimension to raise to
    :param x:
    :param x_len:
    :return:
    """

    original_dim = np.ndim(unc)

    if (original_dim > 2) or (dim > 2):
        raise ValueError(
            "Can only raise uncertainty to a max dimension of 2 (e.g. covariance matrix)"
        )

    if original_dim == dim:
        return unc
    elif unc is None:
        return None
    else:
        if original_dim == 0:
            if (x is None) and (x_len is None):
                raise AttributeError(
                    "Please define either x or x_len to raise dimension of shape 0 uncertainty"
                )
            elif x is not None:
                unc *= x
            else:
                unc = np.full(x_len, unc)

        if dim == 1:
            return unc

        return np.diag(unc)


def func_with_unc(
    f: Callable,
    params: Dict[str, Union[float, np.ndarray]],
    u_params: Dict[str, Union[None, float, np.ndarray]],
    parallel: bool = False,
) -> Tuple[Union[float, np.ndarray], Union[None, float, np.ndarray]]:
    """
    Evaluate function and uncertainties using Monte Carlo method

    Provides simple interface to run a simple Monte Carlo experiment with punpy.

    :param f: function
    :param params: function parameters
    :param u_params: function parameter uncertainties (entries keys same as parameters, one uncertainty per parameter).
    Entries may be of following type:

    * ``None`` - value assumed a constant, no uncertainty (assumed if not provided)
    * ``float`` - assumed to be a relative random uncertainty
    * 1 dimensional ``numpy.ndarray`` - assumed to be an array of absolute random uncertainties (must be same length as
    variable parameter)
    * 2 dimensional ``numpy.ndarray`` - assumed to be a covariance matrix (must be the length of the variable parameter
    square)

    :return: evaluated function and uncertainty
    """

    # evaluate function
    y = f(**params)

    # if no uncertainties return only in band spectrum
    if all(v is None for v in u_params.values()):
        return y, None

    # Add None's for any undefined uncertainties
    u_params_missing = {k: None for k in params.keys() if k not in u_params}
    u_params = {**u_params, **u_params_missing}

    # Find max dimension of uncertainty data and match other variables
    unc_dim = _max_dim(u_params.values())
    if unc_dim != 2:
        unc_dim = 1

    u_params = dict(
        (k, _unc_to_dim(u, unc_dim, x=v)) if (u is not None) else (k, u)
        for (k, u), v in zip(u_params.items(), params.values())
    )

    # Propagate uncertainties
    if parallel:
        prop = MCPropagation(200, parallel_cores=multiprocessing.cpu_count())
    else:
        prop = MCPropagation(200, parallel_cores=1)

    x = [v for v in params.values()]
    u_x = [u_params[k] for k in params.keys()]

    if unc_dim == 1:
        u_y = prop.propagate_random(func=f, x=x, u_x=u_x)
    elif unc_dim == 2:
        u_y = prop.propagate_cov(func=f, x=x, cov_x=u_x, return_corr=False)
    else:
        u_y = None

    del prop

    return y, u_y


if __name__ == "__main__":
    pass



"""
Functions to band integrate spectra for given spectral response function.
"""

import warnings

from matheo.band_integration.srf_utils import (
    return_iter_srf,
    return_band_centres,
    return_band_names,
)
from matheo.utils.punpy_util import func_with_unc
from matheo.utils.function_def import iter_f, f_tophat, f_triangle, f_gaussian
import numpy as np
import scipy.sparse
from typing import Optional, Union, Tuple, List, Iterable, Callable
from comet_maths import interpolate_1d

__author__ = "Sam Hunt"
__created__ = "30/7/2021"


def cutout_nonzero(
    y: np.ndarray,
    x: np.ndarray,
    buffer: Optional[Union[float, int]] = 0.2,
    relative_threshold: Optional[Union[float, int]] = 0.0,
) -> Tuple[np.ndarray, np.ndarray, List[int]]:
    """
    Returns continuous non-zero part of function y(x)

    :param y: function data values
    :param x: function coordinate data values
    :param buffer: fraction of non-zero section of y to include as buffer on either side (default: 0.2)
    :param relative_threshold: (default: 0.) relative threshold to define non-zero region, as fraction of maximum value in y.
    :return: y, x, [imin, imax] - where imin and imax are indices of the start and end of the non-zero region
    """
    # Find extent of non-zero region
    max_val = max(y)
    idx_range = np.zeros(len(y)) * np.nan

    for idx, val in enumerate(y):
        if val > relative_threshold * max_val:
            idx_range[idx] = idx

    idx_range = idx_range[np.isfinite(idx_range)]

    imin = int(min(idx_range))
    imax = int(max(idx_range)) + 1

    # Determine buffer
    width = imax - imin

    imin -= int(width * buffer)
    imax += int(width * buffer)

    imin = imin if imin >= 0 else 0
    imax = imax if imax <= len(y) else len(y)

    return y[imin:imax], x[imin:imax], [imin, imax]


def get_x_offset(y, x, x_centre):
    """
    Returns coordinate offset required to centre function on given position.

    :param y: function values
    :param x: function coordinates
    :param x_centre: function centre defined my location of maximum value.
    :return: coordinate offset to centre function on x_centre
    """

    return x_centre - x[np.argmax(y)]


def _band_int(
    d: np.ndarray,
    x: np.ndarray,
    r: np.ndarray,
    x_r: np.ndarray,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> float:
    """
    Returns integral of data array over a response band (i.e., d(x) * r(x_r))

    N.B.: This function is intended to be wrapped, so it can be applied to an array and run within punpy

    :param d: data to be band integrated
    :param x: data coordinates
    :param r: band response function
    :param x_r: band response function coordinates
    :param rint_norm: (default: True) option to normalise result by integral of r
    :return: band integrated data
    """

    # Cut out non-zero part of SRF to minimise risk of interpolation errors and optimise performance
    r, x_r, idx = cutout_nonzero(r, x_r, **kwargs)

    res_d = (max(x) - min(x)) / len(x)
    res_r = (max(x_r) - min(x_r)) / len(x_r)

    norm_val = 1.0

    # If spectrum lower res than the SRF - interpolate spectrum onto SRF wavelength coordinates before integration
    if res_r < res_d:
        d_interp = interpolate_1d(x, d, x_r)

        if rint_norm:
            norm_val = np.trapz(r, x_r)

        return np.trapz(r * d_interp, x_r) / norm_val

    # If spectrum lower res than the SRF - interpolate spectrum onto SRF wavelength coordinates before integration
    else:
        # First cut out spectrum to SRF wavelength range to avoid extrapolation errors in interpolation
        idx = np.where(np.logical_and(x < max(x_r), x > min(x_r)))
        d = d[idx]
        x = x[idx]

        r_interp = interpolate_1d(x_r, r, x)

        if rint_norm:
            norm_val = np.trapz(r_interp, x)

        return np.trapz(d * r_interp, x) / norm_val


def _band_int_regular_grid(
    d: np.ndarray,
    x: np.ndarray,
    r: np.ndarray,
    d_axis_x: int = 0,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> np.ndarray:
    """
    Returns integral of data array over a response band(s) defined along common, even-spaced coordinates (i.e., d(x) * r(x))

    Accelerated with respect to _band_int using dot product for integral

    N.B.: This function is intended to be wrapped, so it can be run within punpy

    :param d: data to be band integrated
    :param x: data and band response function coordinates along band integration axis, must be evenly spaced
    :param r: band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param d_axis_x: (default 0) x axis in data array
    :param rint_norm: (default: True) option to normalise result by integral of r
    :return: band integrated data
    """

    norm_val = 1.0
    if rint_norm:
        if r.ndim == 1:
            r_sum = np.sum(r)
        else:
            r_sum = np.sum(r, axis=1)

    if d.ndim == 1:
        if rint_norm:
            norm_val = r_sum
        return np.dot(r, d) / norm_val

    dot_dim = 1
    if (r.ndim == 2) and (d.ndim == 2):
        dot_dim = 0

    r_dot_d = np.moveaxis(np.dot(r, np.moveaxis(d, d_axis_x, dot_dim)), 0, d_axis_x)

    if rint_norm:
        sli = [np.newaxis] * d.ndim
        sli[d_axis_x] = slice(None)
        norm_val = r_sum[tuple(sli)]

    return r_dot_d / norm_val


def _band_int_arr(
    d: np.ndarray,
    x: np.ndarray,
    r: np.ndarray,
    x_r: np.ndarray,
    d_axis_x: int = 0,
    rint_norm: bool = False,
    **kwargs: Optional[dict]
) -> np.ndarray:
    """
    Band integrates multi-dimensional data array along x axis.

    In case where ``x == x_r`` and ``x`` is evenly sampled, an accelerated function is used.

    N.B.: This function is intended to be wrapped, so it can be run within punpy

    :param d: data to be band integrated
    :param x: data coordinates along band integration axis
    :param r: band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_r: band response function coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param rint_norm: (default: True) option to normalise result by integral of r
    :return: band integrated data
    """

    # If r and d defined on common regular grid, use accelerated _band_int_regular_grid
    x_sampling = np.diff(x)
    if np.array_equal(x, x_r) and all(x_sampling == x_sampling[0]):
        return _band_int_regular_grid(d, x, r, d_axis_x=d_axis_x, **kwargs)

    # Else run _band_int
    # If single band response function defined, run once
    if r.ndim == 1:

        # If d has multiple dims, ensure integration done along correct axis
        if d.ndim == 1:
            return np.array(
                [_band_int(d, x=x, r=r, x_r=x_r, rint_norm=rint_norm, **kwargs)]
            )
        return np.apply_along_axis(
            _band_int, d_axis_x, arr=d, x=x, r=r, x_r=x_r, rint_norm=rint_norm, **kwargs
        )

    # If multiple band response functions defined, run multiple times
    elif r.ndim == 2:

        # If d has multiple dims, ensure integration done along correct axis
        if d.ndim == 1:
            d_int = np.zeros(len(r))
            for i in range(len(d_int)):
                d_int[i] = _band_int(
                    d, x=x, r=r[i], x_r=x_r, rint_norm=rint_norm, **kwargs
                )

        else:
            # (this bit could probably be accelerated)
            d_int_shape = list(d.shape)
            d_int_shape[d_axis_x] = r.shape[0]
            d_int = np.zeros(d_int_shape)
            for i in range(r.shape[0]):
                # define slice ith band integrated data needs to populate in d_int
                sli = [slice(None)] * d_int.ndim
                sli[d_axis_x] = i
                d_int[tuple(sli)] = np.apply_along_axis(
                    _band_int,
                    d_axis_x,
                    arr=d,
                    x=x,
                    r=r[i],
                    x_r=x_r,
                    rint_norm=rint_norm,
                    **kwargs,
                )

        return d_int


def _band_int2ax_arr(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    rx: np.ndarray,
    x_rx: np.ndarray,
    ry: np.ndarray,
    y_ry: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 1,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> np.ndarray:
    """
    Sequentially band integrates multi-dimensional data array along x axis and y axis

    In case where ``x == x_rx`` and ``x`` is evenly sampled or ``y == y_ry`` and ``y`` is evenly spaced, an accelerated function is used.

    N.B.: This function is intended to be wrapped, so it can be run within punpy

    :param d: data to be band integrated
    :param x: data coordinates along first band integration axis
    :param y: data coordinates along second band integration axis
    :param rx: first axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x_rx``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_rx: first band response function coordinates
    :param ry: second axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``y_ry``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param y_ry: second band response function coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param d_axis_y: (default 1) y axis in data array
    :param rint_norm: (default: True) option to normalise result by integral of r
    :return: band integrated data
    """

    d_intx = _band_int_arr(
        d, x=x, r=rx, x_r=x_rx, d_axis_x=d_axis_x, rint_norm=rint_norm, **kwargs
    )
    d_intx_inty = _band_int_arr(
        d_intx, x=y, r=ry, x_r=y_ry, d_axis_x=d_axis_y, rint_norm=rint_norm, **kwargs
    )

    return d_intx_inty


def _band_int3ax_arr(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    rx: np.ndarray,
    x_rx: np.ndarray,
    ry: np.ndarray,
    y_ry: np.ndarray,
    rz: np.ndarray,
    z_rz: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 1,
    d_axis_z: int = 2,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> np.ndarray:
    """
    Sequentially band integrates multi-dimensional data array along x, y and z axes

    N.B. In case where ``x == x_rx`` and ``x`` is evenly sampled, ``y == y_ry`` and ``y`` is evenly spaced, or ``z == z_rz`` and ``z`` is evenly spaced, an accelerated function is used.

    N.B.: This function is intended to be wrapped, so it can be run within punpy

    :param d: data to be band integrated
    :param x: data coordinates along first band integration axis
    :param y: data coordinates along second band integration axis
    :param z: data coordinates along third band integration axis
    :param rx: first axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x_rx``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_rx: first band response function coordinates
    :param ry: second axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``y_ry``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param y_ry: second band response function coordinates
    :param rz: third axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``z_rz``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param z_rz: third band response function coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param d_axis_y: (default 1) y axis in data array
    :param d_axis_z: (default 2) z axis in data array
    :param rint_norm: (default: True) option to normalise result by integral of r
    :return: band integrated data
    """

    d_intx = _band_int_arr(
        d, x=x, r=rx, x_r=x_rx, d_axis_x=d_axis_x, rint_norm=rint_norm, **kwargs
    )
    d_intx_inty = _band_int_arr(
        d_intx, x=y, r=ry, x_r=y_ry, d_axis_x=d_axis_y, rint_norm=rint_norm, **kwargs
    )
    d_intx_inty_intz = _band_int_arr(
        d_intx_inty,
        x=z,
        r=rz,
        x_r=z_rz,
        d_axis_x=d_axis_z,
        rint_norm=rint_norm,
        **kwargs,
    )

    return d_intx_inty_intz


def band_int(
    d: np.ndarray,
    x: np.ndarray,
    r: np.ndarray,
    x_r: np.ndarray,
    d_axis_x: int = 0,
    x_r_centre: Union[None, float] = None,
    u_d: Union[None, float, np.ndarray] = None,
    u_x: Union[None, float, np.ndarray] = None,
    u_r: Union[None, float, np.ndarray] = None,
    u_x_r: Union[None, float, np.ndarray] = None,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> Union[
    float, np.ndarray, Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
]:
    """
    Returns integral of data array over a response band (i.e., d(x) * r(x_r))

    N.B. In case where ``x == x_r`` and ``x`` is evenly sampled, an accelerated function is used.

    :param d: data to be band integrated
    :param x: data coordinates
    :param r: band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x_r``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_r: band response function coordinates
    :param d_axis_x: (default 0) if d greater than 1D, specify axis to band integrate along
    :param x_r_centre: (optional) centre of band response function in data coordinates, if there is an offset.
     Defined by location of band response function peak in data coordinates.
     Useful to define where sensor is looking along an extended input, e.g. spatially.
    :param u_d: (optional) uncertainty in data
    :param u_x: (optional) uncertainty in data coordinates
    :param u_r: (optional) uncertainty in band response function
    :param u_x_r: (optional) uncertainty in band response function coordinates
    :param rint_norm: (default: True) option to normalise result by integral of r

    :return: band integrated data
    :return: uncertainty of band integrated data (skipped if no input uncertainties provided)
    """

    x_r_off = get_x_offset(r, x_r, x_r_centre) if x_r_centre is not None else 0

    if np.min(x_r + x_r_off) < np.min(x) or np.max(x_r + x_r_off) > np.max(x):
        warnings.warn(
            "The x_r coordinate range (%s-%s) exceeds the range of the x coordinate range (%s-%s), and the band integration can thus not be applied."
            % (np.min(x_r), np.max(x_r), np.min(x), np.max(x))
        )
        return np.nan

    d_band, u_d_band = func_with_unc(
        _band_int_arr,
        params=dict(
            d=d,
            x=x,
            r=r,
            x_r=x_r + x_r_off,
            d_axis_x=d_axis_x,
            rint_norm=rint_norm,
            **kwargs,
        ),
        u_params=dict(d=u_d, x=u_x, r=u_r, x_r=u_x_r),
    )

    if u_d_band is None:
        return d_band

    return d_band, u_d_band


def band_int2ax(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    rx: np.ndarray,
    x_rx: np.ndarray,
    ry: np.ndarray,
    y_ry: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 0,
    x_rx_centre: Union[None, float] = None,
    y_ry_centre: Union[None, float] = None,
    u_d: Union[None, float, np.ndarray] = None,
    u_x: Union[None, float, np.ndarray] = None,
    u_y: Union[None, float, np.ndarray] = None,
    u_rx: Union[None, float, np.ndarray] = None,
    u_x_rx: Union[None, float, np.ndarray] = None,
    u_ry: Union[None, float, np.ndarray] = None,
    u_y_ry: Union[None, float, np.ndarray] = None,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> Union[
    float, np.ndarray, Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
]:
    """
    Sequentially band integrates multi-dimensional data array along x axis and y axis

    N.B. In case where ``x == x_rx`` and ``x`` is evenly sampled or ``y == y_ry`` and ``y`` is evenly spaced, an accelerated function is used.

    :param d: data to be band integrated
    :param x: data coordinates along first band integration axis
    :param y: data coordinates along second band integration axis
    :param rx: first axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x_rx``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_rx: first band response function coordinates
    :param ry: second axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``y_ry``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param y_ry: second band response function coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param d_axis_y: (default 1) y axis in data array
    :param x_rx_centre: (optional) centre of rx function in data coordinates, if there is an offset.
     Defined by location of band response function peak in data coordinates.
     Useful to define where sensor is looking along an extended input, e.g. spatially.
    :param y_ry_centre: (optional) centre of ry function in data coordinates, if there is an offset (as for x_rx_centre)
    :param u_d: (optional) uncertainty in data
    :param u_x: (optional) uncertainty in data coordinates along first band integration axis
    :param u_y: (optional) uncertainty in data coordinates along second band integration axis
    :param u_rx: (optional) uncertainty in first band response function
    :param u_x_rx: (optional) uncertainty in first band response function coordinates
    :param u_ry: (optional) uncertainty in second band response function
    :param u_y_ry: (optional) uncertainty in second band response function coordinates
    :param rint_norm: (default: True) option to normalise result by integral of r

    :return: band integrated data
    :return: uncertainty of band integrated data (skipped if no input uncertainties provided)
    """

    x_rx_off = get_x_offset(rx, x_rx, x_rx_centre) if x_rx_centre is not None else 0
    y_ry_off = get_x_offset(ry, y_ry, y_ry_centre) if y_ry_centre is not None else 0

    d_band, u_d_band = func_with_unc(
        _band_int2ax_arr,
        params=dict(
            d=d,
            x=x,
            y=y,
            rx=rx,
            x_rx=x_rx + x_rx_off,
            ry=ry,
            y_ry=y_ry + y_ry_off,
            d_axis_x=d_axis_x,
            d_axis_y=d_axis_y,
            rint_norm=rint_norm,
            **kwargs,
        ),
        u_params=dict(d=u_d, x=u_x, y=u_y, rx=u_rx, x_rx=u_x_rx, ry=u_ry, y_ry=u_y_ry),
    )

    if u_d_band is None:
        return d_band

    return d_band, u_d_band


def band_int3ax(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    rx: np.ndarray,
    x_rx: np.ndarray,
    ry: np.ndarray,
    y_ry: np.ndarray,
    rz: np.ndarray,
    z_rz: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 1,
    d_axis_z: int = 2,
    x_rx_centre: Union[None, float] = None,
    y_ry_centre: Union[None, float] = None,
    z_rz_centre: Union[None, float] = None,
    u_d: Union[None, float, np.ndarray] = None,
    u_x: Union[None, float, np.ndarray] = None,
    u_y: Union[None, float, np.ndarray] = None,
    u_z: Union[None, float, np.ndarray] = None,
    u_rx: Union[None, float, np.ndarray] = None,
    u_x_rx: Union[None, float, np.ndarray] = None,
    u_ry: Union[None, float, np.ndarray] = None,
    u_y_ry: Union[None, float, np.ndarray] = None,
    u_rz: Union[None, float, np.ndarray] = None,
    u_z_rz: Union[None, float, np.ndarray] = None,
    rint_norm: bool = True,
    **kwargs: Optional[dict]
) -> Union[
    float, np.ndarray, Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
]:
    """
    Sequentially band integrates multi-dimensional data array along x, y and z axes

    N.B. In case where ``x == x_rx`` and ``x`` is evenly sampled, ``y == y_ry`` and ``y`` is evenly spaced, or ``z == z_rz`` and ``z`` is evenly spaced, an accelerated function is used.

    :param d: data to be band integrated
    :param x: data coordinates along first band integration axis
    :param y: data coordinates along second band integration axis
    :param z: data coordinates along third band integration axis
    :param rx: first axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``x_rx``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param x_rx: first band response function coordinates
    :param ry: second axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``y_ry``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param y_ry: second band response function coordinates
    :param rz: third axis band response function(s). For a single band, a 1D length-M array is required, where M is the length of ``z_rz``. Multiple bands may be defined in an N x M array, where N is number of response bands.
    :param z_rz: third band response function coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param d_axis_y: (default 1) y axis in data array
    :param d_axis_z: (default 2) z axis in data array
    :param x_rx_centre: (optional) centre of rx function in data coordinates, if there is an offset.
     Defined by location of band response function peak in data coordinates.
     Useful to define where sensor is looking along an extended input, e.g. spatially.
    :param y_ry_centre: (optional) centre of ry function in data coordinates, if there is an offset (as for x_rx_centre)
    :param z_rz_centre: (optional) centre of rz function in data coordinates, if there is an offset (as for x_rx_centre)
    :param u_d: (optional) uncertainty in data
    :param u_x: (optional) uncertainty in data coordinates along first band integration axis
    :param u_y: (optional) uncertainty in data coordinates along second band integration axis
    :param u_z: (optional) uncertainty in data coordinates along third band integration axis
    :param u_rx: (optional) uncertainty in first band response function
    :param u_x_rx: (optional) uncertainty in first band response function coordinates
    :param u_ry: (optional) uncertainty in second band response function
    :param u_y_ry: (optional) uncertainty in second band response function coordinates
    :param u_rz: (optional) uncertainty in second band response function
    :param u_z_rz: (optional) uncertainty in third band response function coordinates
    :param rint_norm: (default: True) option to normalise result by integral of r

    :return: band integrated data
    :return: uncertainty of band integrated data (skipped if no input uncertainties provided)
    """

    x_rx_off = get_x_offset(rx, x_rx, x_rx_centre) if x_rx_centre is not None else 0
    y_ry_off = get_x_offset(ry, y_ry, y_ry_centre) if y_ry_centre is not None else 0
    z_rz_off = get_x_offset(rz, z_rz, z_rz_centre) if z_rz_centre is not None else 0

    params = dict(
        d=d,
        x=x,
        y=y,
        z=z,
        rx=rx,
        x_rx=x_rx + x_rx_off,
        ry=ry,
        y_ry=y_ry + y_ry_off,
        rz=rz,
        z_ry=z_rz + z_rz_off,
        d_axis_x=d_axis_x,
        d_axis_y=d_axis_y,
        d_axis_z=d_axis_z,
        rint_norm=rint_norm,
        **kwargs,
    )

    u_params = dict(
        d=u_d,
        x=u_x,
        y=u_y,
        z=u_z,
        rx=u_rx,
        x_rx=u_x_rx,
        ry=u_ry,
        y_ry=u_y_ry,
        rz=u_rz,
        z_rz=u_z_rz,
    )

    d_band, u_d_band = func_with_unc(_band_int3ax_arr, params=params, u_params=u_params)

    if u_d_band is None:
        return d_band

    return d_band, u_d_band


def iter_band_int(
    d: np.ndarray,
    x: np.ndarray,
    iter_r: Iterable,
    d_axis_x: int = 0,
    u_d: Optional[Union[float, np.ndarray]] = None,
    u_x: Optional[Union[float, np.ndarray]] = None,
    u_r: Optional[Union[float, np.ndarray]] = None,
    u_x_r: Optional[Union[float, np.ndarray]] = None,
    rint_norm: Optional[bool] = True,
    **kwargs: Optional[dict]
) -> Union[
    float, np.ndarray, Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
]:
    """
    Returns integral of data array over a set of response bands defined by iterator

    :param d: data to be band integrated
    :param x: data coordinates
    :param iter_r: iterable that returns band response function and band response function coordinates at each iteration
    :param d_axis_x: (default 0) if d greater than 1D, specify axis to band integrate along
    :param u_d: uncertainty in data
    :param u_x: uncertainty in data coordinates
    :param u_r: uncertainty in band response function
    :param u_x_r: uncertainty in band response function coordinates
    :param rint_norm: (default: True) option to normalise result by integral of r

    :return: band integrated data
    :return: uncertainty of band integrated data (skipped if no input uncertainties provided)
    """

    # Initialise output data matrix
    d_band_shape = list(d.shape)
    d_band_shape[d_axis_x] = sum(1 for x in iter_r)

    d_band = np.zeros(d_band_shape)
    u_d_band = np.zeros(d_band_shape)

    # Evaluate band integrate
    for i, (r_i, x_r_i) in enumerate(iter_r):

        # define slice ith band integrated data needs to populate in d_band
        sli = [slice(None)] * d_band.ndim
        sli[d_axis_x] = i
        sli = tuple(sli)

        # evaluate band integral
        if (u_d is None) and (u_x is None) and (u_r is None) and (u_x_r is None):
            d_band[sli] = band_int(
                d, x, r_i, x_r_i, d_axis_x, rint_norm=rint_norm, **kwargs
            )
        else:
            d_band[sli], u_d_band[sli] = band_int(
                d,
                x,
                r_i,
                x_r_i,
                d_axis_x,
                u_d=u_d,
                u_x=u_x,
                u_r=u_r,
                u_x_r=u_x_r,
                rint_norm=rint_norm,
                **kwargs,
            )

    if not u_d_band.any():
        return d_band
    else:
        return d_band, u_d_band


def spectral_band_int_sensor(
    d: np.ndarray,
    wl: np.ndarray,
    platform_name: str,
    sensor_name: str,
    detector_name: str = None,
    band_names: Union[None, List[str]] = None,
    d_axis_wl: int = 0,
    u_d: np.ndarray = None,
    u_wl: np.ndarray = None,
    **kwargs: Optional[dict]
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Returns spectral band integrated data array for named sensor spectral bands

    :param d: data to be band integrated
    :param wl: data wavelength coordinates
    :param platform_name: satellite name (must be valid value for
    :param sensor_name: name of instrument on satellite
    :param detector_name: (optional) name of sensor detector. Can be used in sensor has SRF data for for different
    detectors separately - if not specified in this case different
    :param band_names: (optional) sensor bands to evaluate band integral for, if omitted band integral evaluated for
    all bands within spectral range of data
    :param d_axis_wl: (default 0) spectral axis in data array
    :param u_d: (optional) uncertainty in data
    :param u_wl: (optional) uncertainty in data coordinates along first band integration axis

    :return: band integrated data
    :return: band integrated data centre wavelengths
    :return: uncertainties in band integrated data (skipped if no input uncertainties provided)
    """

    # Find bands within data spectral range
    band_names = return_band_names(platform_name, sensor_name, band_names)
    band_centres = return_band_centres(platform_name, sensor_name, band_names)
    valid_idx = np.where(
        np.logical_and(band_centres < max(wl), band_centres > min(wl))
    )[0]
    band_centres = band_centres[valid_idx]
    band_names = [band_names[i] for i in valid_idx]

    # Evaluate band integral
    iter_srf = return_iter_srf(
        platform_name, sensor_name, band_names, detector_name=detector_name
    )

    if (u_d is None) and (u_wl is None):
        return iter_band_int(d, wl, iter_srf, d_axis_wl, **kwargs), band_centres

    d_band, u_d_band = iter_band_int(
        d, wl, iter_srf, d_axis_wl, u_d=u_d, u_x=u_wl, **kwargs
    )
    return d_band, band_centres, u_d_band


def return_r_pixel(
    x_pixel: np.ndarray,
    width_pixel: np.ndarray,
    x: np.ndarray,
    f: Callable,
    x_pixel_off: Optional[float] = None,
) -> np.ndarray:
    """
    Returns per pixel response function, expressed as an n_x X n_pixel matrix, where n_x is the length of wavelength coordinates of the response function defintion and n_pixel matrix is the number of pixels.

    :param x_pixel: centre of band response per pixel
    :param width_pixel: width of band response per pixel
    :param x: coordinates to define band response functions
    :param f: functional shape of response band - a python function with the interface ``f(x, centre, width)``, where ``x`` is a numpy array of the x coordinates to define the function along, ``centre`` is the response band centre, and ``width`` is the response band width.
    :param x_pixel_off: offset to pixel centre locations

    :return: pixel response function matrix
    """

    x_pixel_off = 0.0 if x_pixel_off is None else x_pixel_off

    r_pixel = np.zeros((len(x_pixel), len(x)))
    for i_pixel, (x_pixel_i, width_pixel_i) in enumerate(zip(x_pixel, width_pixel)):
        r_pixel[i_pixel, :] = f(x, x_pixel_i + x_pixel_off, width_pixel_i)

    return r_pixel


def pixel_int(
    d: np.ndarray,
    x: np.ndarray,
    x_pixel: Optional[np.ndarray] = None,
    width_pixel: Optional[np.ndarray] = None,
    u_d: Optional[Union[float, np.ndarray]] = None,
    u_x: Optional[Union[float, np.ndarray]] = None,
    u_x_pixel: Optional[Union[float, np.ndarray]] = None,
    u_width_pixel: Optional[Union[float, np.ndarray]] = None,
    band_shape: Union[Callable, str] = "triangle",
    r_sampling: Optional[float] = None,
    d_axis_x: int = 0,
    x_pixel_centre: Optional[float] = None,
    eval_iter: bool = False,
    **kwargs: Optional[dict]
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Returns integral of data array over a set of response bands (i.e., d(x) * r_i(x_r) for i pixels)

    :param d: data to be band integrated
    :param x: data coordinates
    :param x_pixel: centre of band response per pixel
    :param width_pixel: width of band response per pixel
    :param u_d: uncertainty in data
    :param u_x: uncertainty in data coordinates
    :param u_x_pixel: uncertainty in centre of band response per pixel
    :param u_width_pixel: uncertainty in width of band response per pixel
    :param band_shape: (default: ``triangle``) functional shape of response band - must be either a defined name, one of 'triangle', 'tophat', or 'gaussian', or a python function with the interface ``f(x, centre, width)``, where ``x`` is a numpy array of the x coordinates to define the function along, ``centre`` is the response band centre, and ``width`` is the response band width.
    :param r_sampling: x sampling interval for derived pixel band response functions (if omitted pixel band response functions defined along x, this results in an accelerated computation)
    :param d_axis_x: (default: ``0``) if d greater than 1D, specify axis pixels are along
    :param x_pixel_centre: centre of pixels in data coordinates, if there is an offset.
     Defined as half way between max and min pixel values.
     Useful to define where sensor is looking along an extended input, e.g. spatially.
    :param eval_iter: (default: False) option to evaluate each pixel iteratively, saving memory

    :return: band integrated data
    :return: uncertainty in band integrated data
    """

    # If x_pixel_centre defined compute offset for x_pixel array
    x_pixel_off = (
        x_pixel_centre - ((x_pixel.max() - x_pixel.min()) / 2.0)
        if x_pixel_centre is not None
        else 0
    )

    # Get function
    if band_shape == "triangle":
        f = f_triangle
        xlim_width = 1
    elif band_shape == "tophat":
        f = f_tophat
        xlim_width = 1
    elif band_shape == "gaussian":
        f = f_gaussian
        xlim_width = 3
    else:
        f = band_shape
        xlim_width = 3

    if eval_iter:
        return iter_band_int(
            d,
            x,
            iter_f(f, x_pixel + x_pixel_off, width_pixel, xlim_width=xlim_width),
            d_axis_x,
            u_d,
            u_x,
            **kwargs,
        )

    else:

        # Define r x coordinates
        x_r_pixel = x

        if r_sampling is not None:
            i_x_p_min = np.argmin(x_pixel)
            i_x_p_max = np.argmax(x_pixel)

            x_r_pixel = np.arange(
                x_pixel[i_x_p_min] - xlim_width * width_pixel[i_x_p_min],
                x_pixel[i_x_p_max] + xlim_width * width_pixel[i_x_p_max] + 1,
                r_sampling,
            )

        # Build spectral response function matrix
        r_pixel = return_r_pixel(
            x_pixel,
            width_pixel,
            x_r_pixel,
            f,
            x_pixel_off=x_pixel_off,
        )

        return band_int(d=d, x=x, r=r_pixel, x_r=x_r_pixel, d_axis_x=d_axis_x, **kwargs)


def _band_int2d(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    psf: np.ndarray,
    x_psf: np.ndarray,
    y_psf: np.ndarray,
    **kwargs: Optional[dict]
) -> float:
    """
    Returns integral of a 2D data array over a response band defined by a 2D point spread function
    (i.e., d(x, y) * psf(x_psf, y_psf))

    N.B.: This function is intended to be wrapped, so it can be applied to an array and run within punpy

    :param d: two dimensional data to be band integrated
    :param x: data x coordinates
    :param y: data y coordinates
    :param psf: two dimensional point spread function of band response
    :param x_psf: psf x coordinates
    :param y_psf: psf y coordinates
    :return: band integrated data
    """

    # todo - implement _band_int2d
    raise NotImplementedError


def _band_int2d_arr(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    psf: np.ndarray,
    x_psf: np.ndarray,
    y_psf: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 1,
    **kwargs: Optional[dict]
) -> np.ndarray:
    """
    Integrates two dimensional slice of multi-dimensional data array over a response band defined by a 2D point spread
    function

    N.B.: This function is intended to be wrapped, so it can be run within punpy

    :param d: two dimensional data to be band integrated
    :param x: data x coordinates
    :param y: data y coordinates
    :param psf: two dimensional point spread function of band response
    :param x_psf: psf x coordinates
    :param y_psf: psf y coordinates
    :param d_axis_x: (default 0) x axis in data array
    :param d_axis_y: (default 1) y axis in data array
    :return: band integrated data
    """

    # todo - implement _band_int2d_arr
    raise NotImplementedError


def band_int2d(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    psf: np.ndarray,
    x_psf: np.ndarray,
    y_psf: np.ndarray,
    d_axis_x: int = 0,
    d_axis_y: int = 1,
    u_d: Union[None, float, np.ndarray] = None,
    u_x: Union[None, float, np.ndarray] = None,
    u_y: Union[None, float, np.ndarray] = None,
    u_psf: Union[None, float, np.ndarray] = None,
    u_x_psf: Union[None, float, np.ndarray] = None,
    u_y_psf: Union[None, float, np.ndarray] = None,
    **kwargs: Optional[dict]
) -> Union[
    float, np.ndarray, Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
]:
    """
    Returns integral of a 2D data array over a response band defined by a 2D point spread function
    (i.e., d(x, y) * psf(x_psf, y_psf))

    :param d: two dimensional data to be band integrated
    :param x: data x coordinates
    :param y: data y coordinates
    :param psf: two dimensional point spread function of band response
    :param x_psf: psf x coordinates
    :param y_psf: psf y coordinates
    :param d_axis_x: (default 0) x axis in data array, if d more than 2D
    :param d_axis_y: (default 1) y axis in data array, if d more than 2D
    :param u_d: (optional) uncertainty in data
    :param u_x: (optional) uncertainty in data x coordinates
    :param u_y: (optional) uncertainty in data y coordinates
    :param u_psf: (optional) uncertainty in point spread function of band response
    :param u_x_psf: (optional) uncertainty in psf x coordinates
    :param u_y_psf: (optional) uncertainty in psf x coordinates

    :return: band integrated data
    :return: uncertainty of band integrated data (skipped if no input uncertainties provided)
    """

    d_band, u_d_band = func_with_unc(
        _band_int2d_arr,
        params=dict(
            d=d,
            x=x,
            y=y,
            psf=psf,
            x_psf=x_psf,
            y_psf=y_psf,
            d_axis_x=d_axis_x,
            d_axis_y=d_axis_y,
            **kwargs,
        ),
        u_params=dict(d=u_d, x=u_x, y=u_y, psf=u_psf, x_psf=u_x_psf, y_psf=u_y_psf),
    )

    if u_d_band is None:
        return d_band

    return d_band, u_d_band


def pixel_int2d(
    d: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    x_pixel: np.ndarray,
    y_pixel: np.ndarray,
    width_pixel: np.ndarray,
    psf_shape: str = "triangle",
    d_axis_x: int = 0,
    d_axis_y: int = 0,
    u_d: Union[None, float, np.ndarray] = None,
    u_x: Union[None, float, np.ndarray] = None,
    u_y: Union[None, float, np.ndarray] = None,
    u_x_pixel: Union[None, float, np.ndarray] = None,
    u_y_pixel: Union[None, float, np.ndarray] = None,
    u_width_pixel: Union[None, float, np.ndarray] = None,
    **kwargs: Optional[dict]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns integral of data array over a response band (i.e., d(x) * r(x_r))

    :param d: data to be band integrated
    :param x: data x coordinates
    :param y: data y coordinates
    :param x_pixel: x positions of centre of psf per pixel
    :param y_pixel: y positions of centre of psf per pixel
    :param width_pixel: width of psf per pixel
    :param psf_shape: (default X) psf shape - must be one of...
    :param d_axis_x: (default 0) x axis in data array, if d more than 2D
    :param d_axis_y: (default 1) y axis in data array, if d more than 2D
    :param u_d: uncertainty in data
    :param u_x: uncertainty in data x coordinates
    :param u_y: uncertainty in data y coordinates
    :param u_x_pixel: uncertainty in x positions of centre of psf per pixel
    :param u_y_pixel: uncertainty in y positions of centre of psf per pixel
    :param u_width_pixel: uncertainty in width of psf per pixel

    :return: band integrated data
    :return: uncertainty in band integrated data
    """

    # todo - implement _band_int2d_arr
    raise NotImplementedError


if __name__ == "__main__":
    pass



"""
Tests for band_integration module
"""

import numpy.testing

from matheo.band_integration import band_integration as bi
from matheo.utils import function_def as fd
import unittest
from unittest.mock import patch, call
import numpy as np


"""___Authorship___"""
__author__ = "Sam Hunt"
__created__ = "1/8/2021"


TEST_WL_SPECTRA = np.arange(500, 952, 2)
TEST_SPECTRA = np.array(
    [
        86.40240984,
        87.0470624,
        87.93084099,
        89.34519073,
        88.9919428,
        88.60131038,
        87.64776277,
        84.56509229,
        80.20699595,
        77.57993473,
        79.70188401,
        82.92151658,
        82.80886134,
        80.83847235,
        81.07832761,
        82.60714817,
        81.57882444,
        80.39565515,
        81.47071281,
        81.22992248,
        79.38628691,
        79.1960746,
        80.10704642,
        79.99129846,
        78.99152007,
        78.53258221,
        78.73681595,
        78.59562385,
        77.4069349,
        75.88195983,
        75.29629964,
        75.51243919,
        74.94571446,
        73.74180549,
        73.11458274,
        72.54168798,
        72.24234318,
        71.92226982,
        70.94521004,
        70.45248178,
        70.99364808,
        72.00680476,
        72.3550522,
        71.58434072,
        69.98142462,
        68.92419954,
        70.26385016,
        75.19944433,
        81.83504616,
        84.86159094,
        84.26705631,
        83.90527171,
        84.47465617,
        84.59068069,
        83.82628876,
        82.86556563,
        81.64814889,
        80.20934677,
        79.8305731,
        80.94084382,
        81.73141581,
        80.8522182,
        79.26928993,
        78.39712344,
        78.15093325,
        77.97080866,
        78.13166684,
        78.69433567,
        79.47464896,
        80.23475687,
        80.25169924,
        79.98833192,
        79.82274719,
        79.39371379,
        78.7166021,
        78.48135269,
        79.04004412,
        77.08470515,
        73.92763009,
        75.20355205,
        78.27530866,
        79.20028329,
        78.9060233,
        78.68683697,
        78.68743231,
        78.4101011,
        78.00164631,
        77.85004546,
        77.71481071,
        77.36596503,
        77.02105272,
        76.76404584,
        74.76152893,
        68.14158018,
        64.48681771,
        72.32108895,
        82.55587109,
        87.13105185,
        88.6025905,
        88.77327593,
        87.87524042,
        87.17447792,
        87.54344993,
        87.89064515,
        87.78310067,
        87.4372012,
        86.75809996,
        85.59221129,
        82.45337676,
        77.89214634,
        76.88395498,
        78.8209346,
        79.04027269,
        78.6144584,
        78.56530008,
        78.76360561,
        80.1262072,
        81.5695274,
        81.80366939,
        81.31974949,
        80.66255438,
        80.46491483,
        81.03730472,
        81.40129017,
        81.10318427,
        80.59216682,
        80.49183472,
        80.52269386,
        79.50098152,
        66.05451868,
        40.17258215,
        26.60884751,
        33.40940982,
        50.73591094,
        66.91599287,
        74.31876894,
        76.12143969,
        76.35720283,
        76.37629724,
        76.32542992,
        76.03904999,
        75.60167261,
        75.06019506,
        74.59689624,
        74.14847808,
        73.3689048,
        72.31972623,
        71.67775051,
        71.8884457,
        72.08096512,
        71.6581701,
        71.10508221,
        70.76328245,
        70.59985507,
        70.07171878,
        69.1986223,
        67.97425178,
        65.63276356,
        63.08357537,
        62.156023,
        62.44737329,
        62.65974306,
        63.49386315,
        64.65446038,
        64.57515343,
        63.91961441,
        63.65660967,
        64.05091569,
        64.72807323,
        64.9748043,
        64.97701333,
        64.81060187,
        64.56459521,
        64.51001273,
        63.49677515,
        62.0992306,
        60.46544467,
        58.73146736,
        60.21829619,
        62.54268154,
        62.72533738,
        62.59816901,
        60.95152855,
        58.15937556,
        58.38529592,
        60.43190302,
        61.09648317,
        60.69503418,
        60.39087873,
        60.25081674,
        59.67189389,
        59.29399198,
        59.14432292,
        58.75771598,
        58.81311952,
        58.96741475,
        58.12570394,
        56.19088757,
        53.69796175,
        50.63340289,
        47.91125029,
        48.85782742,
        52.11158853,
        51.94845901,
        49.06416768,
        48.18201855,
        48.92499353,
        48.68125047,
        48.02615653,
        48.55940995,
        49.89487678,
        50.23265632,
        49.75751011,
        49.00514577,
        46.06337159,
        39.22109244,
        31.28268297,
        26.6232523,
        25.62742936,
        28.6281889,
        32.56591759,
        32.48297369,
        31.30941894,
        31.62610506,
        32.32666082,
        32.81016886,
    ]
)


def fake__band_int(d, x, r, x_r, rint_norm=True):
    return 1


def fake_band_int(d, x, r, x_r, d_axis_x, rint_norm=True):
    sli = [slice(None)] * d.ndim
    sli[d_axis_x] = 0
    sli = tuple(sli)

    return np.ones(d[sli].shape)


class FakeBandGen:
    def __init__(self, max_iter=3):
        self.max_iter = max_iter
        self.len_out = 5

    def __iter__(self):

        # Define counter
        self.i = 0
        return self

    def __next__(self):
        """
        Returns ith function

        :return: fake band srf
        :return: band srf wavelength coordinates
        """

        # Iterate through bands
        if self.i < self.max_iter:
            # Update counter
            self.i += 1

            return np.full(self.len_out, self.i), np.arange(self.len_out)

        else:
            raise StopIteration


class TestBandIntegrate(unittest.TestCase):
    def test_cutout_nonzero_buffer(self):
        x = np.arange(20, 80, 0.1)
        y = fd.f_tophat(x, 50, 10)

        x_test = np.arange(43, 57, 0.1)
        y_test = fd.f_tophat(x_test, 50, 10)

        y_eval, x_eval, idx = bi.cutout_nonzero(y, x, buffer=0.2)

        np.testing.assert_array_almost_equal(x_test, x_eval)
        np.testing.assert_array_almost_equal(y_test, y_eval)

    def test_cutout_nonzero_nobuffer(self):
        x = np.arange(20, 80, 0.1)
        y = fd.f_tophat(x, 50, 10)

        x_test = np.arange(45, 55, 0.1)
        y_test = np.ones(x_test.shape)

        y_eval, x_eval, idx = bi.cutout_nonzero(y, x, buffer=0.0)

        np.testing.assert_array_almost_equal(x_test, x_eval)
        np.testing.assert_array_almost_equal(y_test, y_eval)

    def test_get_x_offset(self):
        x = np.arange(0, 40, 1)
        y = fd.f_tophat(x, 20, 10)
        y[23] = 1.1

        x_off = bi.get_x_offset(y, x, 50)

        self.assertEqual(x_off, 27)

    def test__band_int_highressrf(self):

        x = np.arange(0, 100, 0.01)
        d = (0.02 * x) ** 3 + (-0.2 * x) ** 2 + (-3 * x) + 100

        x_r = np.arange(30, 70, 0.001)
        r = fd.f_triangle(x_r, 50, 5)

        x_band = bi._band_int(d, x, r, x_r)

        self.assertAlmostEqual(x_band, 51.1717, places=3)

    def test__band_int_highresspec(self):

        x = np.arange(0, 100, 0.0001)
        d = (0.02 * x) ** 3 + (-0.2 * x) ** 2 + (-3 * x) + 100

        x_r = np.arange(30, 70, 0.001)
        r = fd.f_triangle(x_r, 50, 5)

        x_band = bi._band_int(d, x, r, x_r)

        self.assertAlmostEqual(x_band, 51.1717, places=3)

    def test__band_int_highresspec_nonorm(self):

        x = np.arange(0, 100, 0.0001)
        d = (0.02 * x) ** 3 + (-0.2 * x) ** 2 + (-3 * x) + 100

        x_r = np.arange(30, 70, 0.001)
        r = fd.f_triangle(x_r, 50, 5)

        x_band = bi._band_int(d, x, r, x_r, rint_norm=False)

        self.assertAlmostEqual(x_band, 51.1717 * 5, places=3)

    def test__band_int_regular_grid_r1d_d1d(self):
        d = np.array([4, 4, 4, 4, 4])
        x = np.arange(4)
        r = np.array([1, 1, 1, 1, 1])

        self.assertEqual(bi._band_int_regular_grid(d, x, r), 4)

    def test__band_int_regular_grid_r1d_d1d_nonorm(self):
        d = np.array([4, 4, 4, 4, 4])
        x = np.arange(4)
        r = np.array([1, 1, 1, 1, 1])

        self.assertEqual(bi._band_int_regular_grid(d, x, r, rint_norm=False), 20.0)

    def test__band_int_regular_grid_r2d_d2d(self):
        d = np.array(
            [[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3]],
        )

        x = np.arange(4)

        r = np.array(
            [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1]]
        )

        d_int = bi._band_int_regular_grid(d, x, r, d_axis_x=1)

        d_int_expected = np.array(
            [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]],
        )

        np.testing.assert_array_equal(d_int, d_int_expected)

    def test__band_int_regular_grid_r2d_d3d(self):
        d = np.array(
            [
                [[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3]],
                [[4, 4, 4, 4, 4], [5, 5, 5, 5, 5], [6, 6, 6, 6, 6]],
            ]
        )

        x = np.arange(4)

        r = np.array(
            [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1]]
        )

        d_int = bi._band_int_regular_grid(d, x, r, d_axis_x=2)

        d_int_expected = np.array(
            [
                [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]],
                [[4, 4, 4, 4], [5, 5, 5, 5], [6, 6, 6, 6]],
            ]
        )

        np.testing.assert_array_equal(d_int, d_int_expected)

    @patch("matheo.band_integration.band_integration._band_int_regular_grid")
    def test__band_int_arr_regular_grid(self, mock):
        x = np.arange(4)
        x_r = np.arange(4)
        d_band = bi._band_int_arr("d", x, "r", x_r, d_axis_x=2)

        mock.assert_called_once_with("d", x, "r", d_axis_x=2)

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int_arr_r1d_d1d(self, mock):
        d = np.zeros(5)
        x = np.arange(5)
        x_r = np.arange(10)
        r = fd.f_triangle(x_r, 5, 5)

        d_band = bi._band_int_arr(d, x, r, x_r)

        np.testing.assert_array_equal(np.ones(1), d_band)

        self.assertEqual(1, mock.call_count)

        np.testing.assert_array_equal(mock.call_args_list[0][0][0], d)
        np.testing.assert_array_equal(mock.call_args_list[0][1]["x"], x)
        np.testing.assert_array_equal(mock.call_args_list[0][1]["r"], r)
        np.testing.assert_array_equal(mock.call_args_list[0][1]["x_r"], x_r)

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int_arr_r1d_d3d(self, mock):

        d = np.zeros((3, 4, 5))
        x = np.arange(5)
        x_r = np.arange(10)
        r = fd.f_triangle(x_r, 5, 5)

        d_band = bi._band_int_arr(d, x, r, x_r, d_axis_x=2)

        np.testing.assert_array_equal(np.ones((3, 4)), d_band)

        self.assertEqual(12, mock.call_count)

        expected_calls = []
        for i in range(12):
            expected_calls.append(call(np.zeros(5), x=x, r=r, x_r=x_r))

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])

            real_kwargs = real_call[1]
            expected_kwargs = expected_call[2]
            np.testing.assert_array_equal(real_kwargs["x"], expected_kwargs["x"])
            np.testing.assert_array_equal(real_kwargs["r"], expected_kwargs["r"])
            np.testing.assert_array_equal(real_kwargs["x_r"], expected_kwargs["x_r"])

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int_arr_r2d_d1d(self, mock):
        d = np.zeros(5)
        x = np.arange(5)
        x_r = np.arange(10)
        r = np.zeros((5, 10))
        r[0, :] = 0.0
        r[1, :] = 1.0
        r[2, :] = 2.0
        r[3, :] = 3.0
        r[4, :] = 4.0

        d_band = bi._band_int_arr(d, x, r, x_r, d_axis_x=2)

        np.testing.assert_array_equal(np.ones(5), d_band)

        self.assertEqual(5, mock.call_count)

        expected_calls = []
        for i in range(5):
            expected_calls.append(call(np.zeros(5), x=x, r=np.full(10, i), x_r=x_r))

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])

            real_kwargs = real_call[1]
            expected_kwargs = expected_call[2]
            np.testing.assert_array_equal(real_kwargs["x"], expected_kwargs["x"])
            np.testing.assert_array_equal(real_kwargs["r"], expected_kwargs["r"])
            np.testing.assert_array_equal(real_kwargs["x_r"], expected_kwargs["x_r"])

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int_arr_r2d_d3d(self, mock):
        d = np.zeros((3, 4, 5))
        x = np.arange(5)
        x_r = np.arange(10)
        r = np.zeros((6, 10))
        r[0, :] = 0.0
        r[1, :] = 1.0
        r[2, :] = 2.0
        r[3, :] = 3.0
        r[4, :] = 4.0
        r[5, :] = 5.0

        d_band = bi._band_int_arr(d, x, r, x_r, d_axis_x=2)

        np.testing.assert_array_equal(np.ones((3, 4, 6)), d_band)

        self.assertEqual(72, mock.call_count)

        expected_calls = []
        for i in range(5):
            for j in range(12):
                expected_calls.append(call(np.zeros(5), x=x, r=np.full(10, i), x_r=x_r))

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])

            real_kwargs = real_call[1]
            expected_kwargs = expected_call[2]
            np.testing.assert_array_equal(real_kwargs["x"], expected_kwargs["x"])
            np.testing.assert_array_equal(real_kwargs["r"], expected_kwargs["r"])
            np.testing.assert_array_equal(real_kwargs["x_r"], expected_kwargs["x_r"])

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int2ax_arr(self, mock):

        d = np.zeros((3, 4, 5))
        x = np.arange(5)
        y = np.arange(3)
        x_rx = np.arange(10)
        rx = fd.f_triangle(x_rx, 5, 5)
        y_ry = np.arange(5)
        ry = fd.f_triangle(y_ry, 5, 3)

        d_band = bi._band_int2ax_arr(
            d, x, y, rx, x_rx, ry, y_ry, d_axis_x=2, d_axis_y=0
        )

        np.testing.assert_array_equal(np.ones(4), d_band)

        self.assertEqual(16, mock.call_count)

        expected_calls = []
        for i in range(12):
            expected_calls.append(call(np.zeros(5), x=x, r=rx, x_r=x_rx))

        for i in range(4):
            expected_calls.append(call(np.ones(3), x=y, r=ry, x_r=y_ry))

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])

            real_kwargs = real_call[1]
            expected_kwargs = expected_call[2]
            np.testing.assert_array_equal(real_kwargs["x"], expected_kwargs["x"])
            np.testing.assert_array_equal(real_kwargs["r"], expected_kwargs["r"])
            np.testing.assert_array_equal(real_kwargs["x_r"], expected_kwargs["x_r"])

    @patch("matheo.band_integration.band_integration._band_int", wraps=fake__band_int)
    def test_band_int3ax_arr(self, mock):

        d = np.zeros((3, 4, 5))
        x = np.arange(5)
        y = np.arange(3)
        z = np.arange(4)
        x_rx = np.arange(10)
        rx = fd.f_triangle(x_rx, 5, 5)
        y_ry = np.arange(5)
        ry = fd.f_triangle(y_ry, 5, 3)
        z_rz = np.arange(7)
        rz = fd.f_triangle(y_ry, 5, 3)

        d_band = bi._band_int3ax_arr(
            d, x, y, z, rx, x_rx, ry, y_ry, rz, z_rz, d_axis_x=2, d_axis_y=0, d_axis_z=1
        )

        np.testing.assert_array_equal(np.array([1]), d_band)

        self.assertEqual(17, mock.call_count)

        expected_calls = []
        for i in range(12):
            expected_calls.append(call(np.zeros(5), x=x, r=rx, x_r=x_rx))

        for i in range(4):
            expected_calls.append(call(np.ones(3), x=y, r=ry, x_r=y_ry))

        expected_calls.append(call(np.ones(4), x=z, r=rz, x_r=z_rz))

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])

            real_kwargs = real_call[1]
            expected_kwargs = expected_call[2]
            np.testing.assert_array_equal(real_kwargs["x"], expected_kwargs["x"])
            np.testing.assert_array_equal(real_kwargs["r"], expected_kwargs["r"])
            np.testing.assert_array_equal(real_kwargs["x_r"], expected_kwargs["x_r"])

    @patch("matheo.band_integration.band_integration.band_int", wraps=fake_band_int)
    def test_iter_band_int(self, mock):
        d = np.zeros((3, 4, 11))
        x = np.arange(11)
        fakebandgen = FakeBandGen(2)
        fakebanditer = iter(fakebandgen)

        d_band = bi.iter_band_int(d, x, fakebanditer, d_axis_x=2)

        self.assertEqual(d_band.shape, (3, 4, 2))
        np.testing.assert_array_equal(d_band, np.ones(d_band.shape))

        expected_calls = [
            call(d, x, np.full(5, 1), np.arange(5), 2),
            call(d, x, np.full(5, 2), np.arange(5), 2),
        ]

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])
            np.testing.assert_array_equal(real_call[0][1], expected_call[1][1])
            np.testing.assert_array_equal(real_call[0][2], expected_call[1][2])
            np.testing.assert_array_equal(real_call[0][3], expected_call[1][3])
            self.assertEqual(real_call[0][4], expected_call[1][4])

    @patch("matheo.band_integration.band_integration.band_int", wraps=fake_band_int)
    def test_spectral_band_int_sensor(self, mock):
        d = np.zeros((3, 4, 11))
        wl = np.arange(400, 510, 10)

        d_band, wl_band = bi.spectral_band_int_sensor(
            d,
            wl,
            d_axis_wl=2,
            platform_name="Sentinel-2A",
            sensor_name="msi",
        )

        self.assertEqual(d_band.shape, (3, 4, 2))
        np.testing.assert_array_equal(d_band, np.ones(d_band.shape))

        expected_calls = [
            call(
                d,
                wl,
                np.array(
                    [
                        0.00000000e00,
                        0.00000000e00,
                        1.77574169e-03,
                        4.07306058e-03,
                        3.62614286e-03,
                        3.51519883e-03,
                        5.72916260e-03,
                        3.78029188e-03,
                        2.63673207e-03,
                        1.26211275e-03,
                        1.98758300e-03,
                        1.36891310e-03,
                        1.25044398e-03,
                        4.63454402e-04,
                        8.14292987e-04,
                        1.37643155e-03,
                        1.48508593e-03,
                        1.82373472e-03,
                        1.62681751e-03,
                        4.39206185e-03,
                        2.90080979e-02,
                        1.18745930e-01,
                        3.23875070e-01,
                        5.72819233e-01,
                        7.14727521e-01,
                        7.61967778e-01,
                        7.89297044e-01,
                        8.08623850e-01,
                        8.10893834e-01,
                        8.24198782e-01,
                        8.54158103e-01,
                        8.70790899e-01,
                        8.87310982e-01,
                        9.26199257e-01,
                        9.82281506e-01,
                        1.00000000e00,
                        9.75238204e-01,
                        9.35963392e-01,
                        8.89971495e-01,
                        8.50210488e-01,
                        8.25694501e-01,
                        7.83902407e-01,
                        6.14174187e-01,
                        3.30071092e-01,
                        1.24108315e-01,
                        4.36569415e-02,
                        1.47495950e-02,
                        0.00000000e00,
                        0.00000000e00,
                    ]
                ),
                np.arange(410, 459, 1),
                2,
            ),
            call(
                d,
                wl,
                np.array(
                    [
                        0.0000000e00,
                        0.0000000e00,
                        4.2555310e-02,
                        7.2298303e-02,
                        1.5374322e-01,
                        3.2799226e-01,
                        5.5336785e-01,
                        7.1011168e-01,
                        7.5285178e-01,
                        7.5232691e-01,
                        7.5668079e-01,
                        7.6326948e-01,
                        7.6239425e-01,
                        7.8525150e-01,
                        8.1546670e-01,
                        8.6179179e-01,
                        8.9282596e-01,
                        9.1952211e-01,
                        9.1900647e-01,
                        9.1315752e-01,
                        9.0035367e-01,
                        8.8989693e-01,
                        8.8232458e-01,
                        8.7606120e-01,
                        8.8429987e-01,
                        9.0695542e-01,
                        9.3232083e-01,
                        9.3947250e-01,
                        9.4383544e-01,
                        9.2204088e-01,
                        8.8602310e-01,
                        8.4743607e-01,
                        8.1251687e-01,
                        7.8239709e-01,
                        7.7310872e-01,
                        7.7209055e-01,
                        7.8742653e-01,
                        8.1217176e-01,
                        8.4605050e-01,
                        8.8767993e-01,
                        9.2793995e-01,
                        9.5069236e-01,
                        9.6573311e-01,
                        9.6938252e-01,
                        9.6570295e-01,
                        9.5832002e-01,
                        9.5405066e-01,
                        9.5178270e-01,
                        9.5699722e-01,
                        9.6556515e-01,
                        9.7705138e-01,
                        9.7709572e-01,
                        9.7436607e-01,
                        9.5903182e-01,
                        9.3506318e-01,
                        9.0190136e-01,
                        8.7165791e-01,
                        8.4402442e-01,
                        8.2280850e-01,
                        8.1536043e-01,
                        8.2057637e-01,
                        8.3951491e-01,
                        8.6992168e-01,
                        9.1526204e-01,
                        9.6067029e-01,
                        9.9163699e-01,
                        1.0000000e00,
                        9.8356098e-01,
                        9.1130763e-01,
                        7.4018258e-01,
                        5.0395858e-01,
                        3.0501550e-01,
                        1.8004605e-01,
                        1.0738342e-01,
                        6.5935917e-02,
                        4.2077459e-02,
                        2.6621290e-02,
                        1.4339600e-02,
                        2.6577900e-03,
                        8.1822003e-04,
                        0.0000000e00,
                        0.0000000e00,
                    ]
                ),
                np.arange(454, 536, 1),
                2,
            ),
        ]

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])
            np.testing.assert_array_equal(real_call[0][1], expected_call[1][1])
            np.testing.assert_array_almost_equal(real_call[0][2], expected_call[1][2])
            np.testing.assert_array_almost_equal(
                real_call[0][3], expected_call[1][3], decimal=4
            )
            self.assertEqual(real_call[0][4], expected_call[1][4])

    @patch("matheo.band_integration.band_integration.return_r_pixel")
    @patch("matheo.band_integration.band_integration.band_int")
    def test_pixel_int(self, mock_bi, mock_rrp):
        d = np.zeros(12)
        x = np.arange(12)
        x_pixel = np.array([5, 10])
        width_pixel = np.array([2, 4])
        d_axis_x = 0

        d_band = bi.pixel_int(
            d=d, x=x, x_pixel=x_pixel, width_pixel=width_pixel, d_axis_x=d_axis_x
        )

        mock_bi.assert_called_once_with(
            d=d, x=x, r=mock_rrp.return_value, x_r=x, d_axis_x=d_axis_x
        )

    @patch("matheo.band_integration.band_integration.return_r_pixel")
    @patch("matheo.band_integration.band_integration.band_int")
    def test_pixel_int_r_sampling(self, mock_bi, mock_rrp):
        d = np.zeros(12)
        x = np.arange(12)
        x_pixel = np.array([5, 10])
        width_pixel = np.array([2, 4])
        d_axis_x = 0

        d_band = bi.pixel_int(
            d=d, x=x, x_pixel=x_pixel, width_pixel=width_pixel, d_axis_x=d_axis_x
        )

        mock_bi.assert_called_once_with(
            d=d, x=x, r=mock_rrp.return_value, x_r=x, d_axis_x=d_axis_x
        )

    @patch("matheo.band_integration.band_integration.return_r_pixel")
    @patch("matheo.band_integration.band_integration.band_int")
    def test_pixel_int_r_sampling(self, mock_bi, mock_rrp):
        d = np.zeros(12)
        x = np.arange(12)
        x_pixel = np.array([5, 10])
        width_pixel = np.array([2, 4])
        d_axis_x = 0

        d_band = bi.pixel_int(
            d=d,
            x=x,
            x_pixel=x_pixel,
            width_pixel=width_pixel,
            r_sampling=1.0,
            band_shape="tophat",
            d_axis_x=d_axis_x,
        )

        x_r = np.arange(3.0, 15.0)

        np.testing.assert_array_almost_equal(mock_bi.call_args[1]["d"], d)
        np.testing.assert_array_almost_equal(mock_bi.call_args[1]["x"], x)
        self.assertEqual(mock_bi.call_args[1]["r"], mock_rrp.return_value)
        np.testing.assert_array_almost_equal(mock_bi.call_args[1]["x_r"], x_r)
        self.assertEqual(mock_bi.call_args[1]["d_axis_x"], d_axis_x)

    @patch("matheo.band_integration.band_integration.band_int", wraps=fake_band_int)
    def test_pixel_int_eval_iter(self, mock):
        d = np.zeros((3, 4, 11))
        x = np.arange(11)
        x_pixel = np.array([5, 10])
        width_pixel = np.array([2, 4])

        d_band = bi.pixel_int(
            d=d,
            x=x,
            x_pixel=x_pixel,
            width_pixel=width_pixel,
            d_axis_x=2,
            eval_iter=True,
        )

        self.assertEqual(d_band.shape, (3, 4, 2))
        np.testing.assert_array_equal(d_band, np.ones(d_band.shape))

        x_r_0 = np.arange(
            5 - 2, 5 + 2 + 1, 0.01
        )  # (centre-width, centre+width+1, 0.01)
        # r_0 = f_triangle(x_r_0, centre, width):
        r_0 = np.zeros(x_r_0.shape)
        first_half = np.logical_and(5 - 2 < x_r_0, x_r_0 <= 5)
        r_0[first_half] = (x_r_0[first_half] - (5 - 2)) / (x_pixel[0] - (5 - 2))

        second_half = np.logical_and(x_pixel[0] < x_r_0, x_r_0 < (5 + 2))
        r_0[second_half] = ((5 + 2) - x_r_0[second_half]) / ((5 + 2) - 5)

        x_r_1 = np.arange(
            10 - 4, 10 + 4 + 1, 0.01
        )  # (centre-width, centre+width+1, 0.01)
        # r_1 = f_triangle(x_r_1, centre, width):
        r_1 = np.zeros(x_r_1.shape)
        first_half = np.logical_and((10 - 4) < x_r_1, x_r_1 <= 10)
        r_1[first_half] = (x_r_1[first_half] - (10 - 4)) / (10 - (10 - 4))

        second_half = np.logical_and(x_pixel[1] < x_r_1, x_r_1 < (10 + 4))
        r_1[second_half] = ((10 + 4) - x_r_1[second_half]) / ((10 + 4) - 10)

        expected_calls = [call(d, x, r_0, x_r_0, 2), call(d, x, r_1, x_r_1, 2)]

        for expected_call, real_call in zip(expected_calls, mock.call_args_list):
            np.testing.assert_array_equal(real_call[0][0], expected_call[1][0])
            np.testing.assert_array_equal(real_call[0][1], expected_call[1][1])
            np.testing.assert_array_equal(real_call[0][2], expected_call[1][2])
            np.testing.assert_array_equal(real_call[0][3], expected_call[1][3])
            self.assertEqual(real_call[0][4], expected_call[1][4])

    def test_return_r_pixel(self):

        x = np.arange(20)
        x_pixel = np.array([4.0, 8.0, 12.0])
        width_pixel = np.array([2.0, 2.0, 4.0])
        r_pixel = bi.return_r_pixel(x_pixel, width_pixel, x, fd.f_tophat)

        expected_r_pixel = np.zeros((3, 20))

        expected_r_pixel[0, 3:6] = 1.0
        expected_r_pixel[1, 7:10] = 1.0
        expected_r_pixel[2, 10:15] = 1.0

        numpy.testing.assert_almost_equal(r_pixel, expected_r_pixel)


if __name__ == "__main__":
    unittest.main()



import io
import os
import re

from setuptools import find_packages
from setuptools import setup

exec(open("matheo/_version.py").read())


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type("")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    version=__version__,
    name="matheo",
    url="https://github.com/meteor-toolkit/matheo",
    license="None",
    author="Sam Hunt, Pieter De Vis",
    author_email="sam.hunt@npl.co.uk",
    description="Matheo is a python package with mathematical algorithms for use in Earth observation data and tools",
    long_description=read("README.md"),
    packages=find_packages(exclude=("tests",)),
    install_requires=["pyspectral", "comet_maths>=1.0.5", "punpy>=1.0.4"],
    extras_require={
        "dev": [
            "pre-commit",
            "tox",
            "sphinx",
            "sphinx_design",
            "sphinx_book_theme",
            "ipython",
            "sphinx_autosummary_accessors",
        ],
    },
)
